// A poly data type demo program. Theorem proving.
// Compare with Appendix A in "Chang C.L and Lee R.C. Symbolic
// Logic and Mechanical Theorem Proving", Academic Press, New York, 1973

program TheoremProving;
const
  typeSTR  = GetTypeID('String');
  typeCHAR = GetTypeID('Char');
type
  TClause = class
            private
              E, Param: Poly;
              Number, NParam, NX, NY, ND: Integer;
              function Copy(const A: Poly): Poly;
              function GetCount: Integer;
            public
              constructor Create(const A: Poly);
              constructor Resolvent(const A, B: TClause; LA, LB: Integer);
              destructor Free;
              procedure Rename(const L: Poly);
              procedure Print;
            end;

  TCollect = class
               Items : array of TClause;
               Card : Integer;
               constructor Create;
               destructor Free;
               procedure Add(C: TClause);
               procedure AddCollection(Coll: TCollect);
               procedure Print;
             end;

  THistoryRec = record
                  N1, N2, ND: Integer;
                  InProof: boolean;
                end;

var
  ArrConst, AX, AY: Poly; // logical variables and constants

  Num : Integer; // clause counter

  HIST: array[1..1000] of THistoryRec;
  HC  : Integer;  // length of history

function IsLogVar(const P: Poly): boolean;
begin
  result := (PolyType(P) = typeSTR) and (Copy(String(P), 1, 1) = '#');
end;

function IsCompound(const P: Poly): boolean;
begin
  result := Length(P) > 0;
end;

function Inside(const Key, T: Poly): boolean;
var
  I: Integer;
begin
  if IsCompound(T) then
    for I:=2 to Length(T) do
    begin
      result := Inside(Key, T[I]);
      if result then
        Exit;
    end
  else
    result := Key = T;
end;

function Unify(N: Integer; const T1, T2: Poly): boolean;
var
  I: Integer;
  P1, P2: Poly;
begin
  result := true;
  for I:=N to Length(T1) do
  begin
    P1 == T1[I];
    P2 == T2[I];
    if IsCompound(P1) and IsCompound(P2) then
      result := Unify(1, P1, P2)
    else if IsLogVar(P1) and (not Inside(P1, P2)) then
    begin
      deallocate P1^;
      P1^ == P2^;
    end
    else if IsLogVar(P2) and (not Inside(P2, P1)) then
    begin
      deallocate P2^;
      P2^ == P1^;
    end
    else
      result := P1 = P2;
    if not result then
      Exit;
  end;
end;

function IsVar(const P: Poly): boolean;
var
  C : Char;
begin
  result := PolyType(P) = typeCHAR;
  if not result then Exit;
  C := Char(P);
  result := (C >= 'U') and (C <= 'Z');
end;

function IsAnswerPredicate(const P: Poly): boolean;
var
  T: Integer;
begin
  result := false;
  T := PolyType(P);
  if T <> typeSTR then Exit;
  result := String(P) = 'ans';
end;

constructor TClause.Create(const A: Poly);

var
  UsedVars: array[1..10] of Char;

procedure SetVars(const P: Poly); 
var
  I, J, K: Integer;
  C: Char;
  PI: Poly; 
begin
  for I:=1 to Length(P) do
  begin
    PI == P[I]; 
    if IsCompound(PI) then
      SetVars(PI)
    else if IsVar(PI) then
    begin
      C := Char(PI); 
      K := 0; 
      for J:=1 to NParam do
        if UsedVars[J] = C then
          K := J; 
      if K = 0 then
      begin
        Inc(NParam);
        K := NParam;
        UsedVars[K] := C;
      end;
      Param[K] := AX[K];
      deallocate PI^;
      PI^ == Param[K];
    end; 
  end;
end;

begin
  SetLength(Param, 10); 
  NParam := 0; 
  E := A;
  SetVars(E);
  Inc(Num);
  Number := Num;
  NX     := 1000;
  NY     := 0;
  ND     := 0;
end;

destructor TClause.Free;
begin
  deallocate E, Param; 
end;

procedure TClause.Rename(const L: Poly);
var
  I: Integer;
begin
  for I:=1 to NParam do
  begin
    deallocate Param[I];
    Param[I] := L[I];
  end;
end;

procedure TClause.Print;
begin
  writeln(Number,'(',NX,',',NY,',',ND,') E=',E);
end;

function TClause.Copy(const A: Poly): Poly;
var
  I, J, N: Integer;
  AI: Poly;
  Exist: boolean;
begin
  N := Length(A);
  SetLength(result, N);
  for I:=1 to N do
  begin
    AI == A[I];
    if      IsCompound(AI) then
      result[I] := Copy(AI)
    else if IsLogVar(AI) then
    begin
      Exist := false;
      for J:=1 to NParam do
        if Param[J] = AI then
        begin
          Exist := true;
          result[I] == Param[J];
        end;
      if not Exist then
      begin
        Inc(NParam);
        Param[NParam] := AI;
        result[I] == Param[NParam];
      end
    end
    else
      result[I] := AI;
  end;
end;

function TClause.GetCount: Integer;
begin
  result := Length(E);
  while (result > 0) and IsAnswerPredicate(E[result, 2]) do
    Dec(result);
end;

constructor TClause.Resolvent(const A, B: TClause; LA, LB: Integer);
var
  N, N1, N2, I, J: Integer;
begin
  N1 := Length(A.E);
  N2 := Length(B.E);
  N  := N1 + N2 - 2;
  SetLength(E, N);
  SetLength(Param, 10); 
  NParam := 0;

  J := 0;
  for I:=1 to A.GetCount() do                                        
    if I <> LA then
    begin
      Inc(J);
      E[J] := Copy(A.E[I]);
    end;

  for I:=1 to B.GetCount() do                                        
    if I <> LB then
    begin
      Inc(J);
      E[J] := Copy(B.E[I]);
    end;

  while (N1 > 0) and IsAnswerPredicate(A.E[N1, 2]) do 
  begin
    Inc(J);
    E[J] := Copy(A.E[N1]);
    Dec(N1);
  end;

  while (N2 > 0) and IsAnswerPredicate(B.E[N2, 2]) do
  begin 
    Inc(J);
    E[J] := Copy(B.E[N2]);
    Dec(N2);
  end;

  Inc(Num);
  Number := Num;
  NX     := A.Number;
  NY     := B.Number;
  ND     := LB;
end;

constructor TCollect.Create;
begin
  SetLength(Items, 4);
  Card := 0; 
end;

destructor TCollect.Free;
begin
  SetLength(Items, 0);
end;

procedure TCollect.Add(C: TClause);
begin
  if Card = Length(Items) then
    SetLength(Items, Length(Items) + 10);
  Items[Card] := C;
  Inc(Card);
end;

procedure TCollect.AddCollection(Coll: TCollect); 
var
  I: Integer;
begin
  for I:=0 to Coll.Card-1 do
    Add(Coll.Items[I]);
end;

procedure TCollect.Print;
var
  I: Integer;
begin
  writeln;
  writeln('Collection, Card=',Card);
  for I:=0 to Card-1 do
    Items[I].Print;
end;

function Contradict(U, V: TCollect): boolean;
var
  I, J: Integer;
  UI, VJ: TClause;
  Functor: String;
begin
  result := false;
  for I:=0 to U.Card-1 do
  begin
    UI := U.Items[I];
    Functor := String(UI.E[1, 2]);
    for J:=0 to V.Card-1 do
    begin
      VJ := V.Items[J];
      if Functor = String(VJ.E[1, 2]) then
      begin
        UI.Rename(AX);
        VJ.Rename(AY);
        result := Unify(3, UI.E[1], VJ.E[1]);
        if result then
        begin
          Inc(HC);
          HIST[HC].N1 := UI.Number;
          HIST[HC].N2 := VJ.Number;
          Exit;
        end;
      end;
    end;
  end;
end;

function STest(S: TCollect; U: TClause): boolean;
var
  Functor: String;
  I: Integer;
  SI: TClause;
begin
  result := false;
  Functor := String(U.E[1, 2]);
  for I:=0 to S.Card-1 do
  begin
    SI := S.Items[I];
    if Functor = String(SI.E[1, 2]) then
    begin
      SI.Rename(ArrConst);
      U.Rename(AX);
      result := Unify(3, SI.E[1], U.E[1]);
      if result then
        Exit;
    end;
  end;
end;

function GUnit(var A: array of TClause; var LA: Integer;
               S1, S2, W: TCollect;
               C: TClause): boolean;
var
  P, S, WW, POS, NEG: TCollect;
  Stack: array[1..100] of TClause;
  I, J, NS: Integer;
  U, R: TClause;
  SU, FU: String;
begin
  WW  := TCollect.Create;
  POS := TCollect.Create;
  NEG := TCollect.Create;

  try
    WW.AddCollection(S1);
    WW.AddCollection(S2);

    Stack[1] := C;
    NS := 1;

    repeat
      C := Stack[NS];
      Dec(NS);
      for I:=0 to W.Card-1 do
      begin
        U := W.Items[I];
        if U.Number < C.NX then
        begin
          SU := String(U.E[1, 1]);
          FU := String(U.E[1, 2]);
          for J:=1 to C.GetCount() do
            if (SU <> String(C.E[J, 1])) and (FU = String(C.E[J, 2])) then
            begin
              U.Rename(AX);
              C.Rename(AY);
              if Unify(3, U.E[1], C.E[J]) then
              begin
                R := TClause.Resolvent(U, C, 1, J);
                
                writeln('Resolvent:');
                R.Rename(AX);
                R.Print();

                Inc(HC);
                HIST[HC].N1 := U.Number;
                HIST[HC].N2 := C.Number;
                HIST[HC].ND := J;

                if R.GetCount() = 1 then
                begin
                  if String(R.E[1, 1]) = '+' then
                  begin
                    S := S1;
                    P := POS;
                  end
                  else
                  begin
                    S := S2;
                    P := NEG;
                  end;
                  if (not STest(S, R)) and (not STest(P, R)) then
                  begin
                    Inc(LA);
                    A[LA] := R;
                    P.Add(R);
                  end;
                end
                else
                begin
                  Inc(NS);
                  Stack[NS] := R;
                end;
              end;
            end;
        end;
      end;
      W := WW;
    until NS = 0;

    result := Contradict(S1, NEG);
    if not result then
    begin
      result := Contradict(S2, POS);
      if not result then
      begin
        S1.AddCollection(POS);
        S2.AddCollection(NEG);
      end;
    end;
  finally
    POS.Free;
    NEG.Free;
    WW.Free;
  end;
end;

procedure MarkTree(I, N: Integer);
begin
  if I > N then
  begin
    HIST[I].InProof := true;
    MarkTree( HIST[I].N1, N);
    MarkTree( HIST[I].N2, N);
  end;
end;

function TPU(var A: array of TClause; 
             SupportSet: Poly; 
             N1, N2: Integer): boolean;
var
  S, S1, S2, S3: TCollect;
  I, J, K, L, LA, N: Integer;
  W: array of TCollect;
  C: TClause;
begin
  S1 := TCollect.Create;
  S2 := TCollect.Create;
  S3 := TCollect.Create;

  try
    HC := N1;
    SetLength(W, N1);

    LA := N1;

    for I:=1 to N1 do
    begin
      C := A[I];
      if C.GetCount() = 1 then
      begin
        if String(C.E[1, 1]) = '+' then
          S := S1                         // positive clause
        else
          S := S2;                        // negative clause
      end
      else
        S := S3;                          // non-unit clause
      S.Add(C);

      if Assigned(SupportSet[I]) then
      begin
        K := S3.Card - 1;
        W[K] := TCollect.Create;
        for J:=1 to Length(SupportSet[I]) do
          W[K].Add(A[Integer(SupportSet[I,J])]);
      end;
    end;

    result := Contradict(S1, S2);

    if result then
      Exit;

    K := 0;

    for I:=1 to N2 do
    begin
      N := LA;
      C := S3.Items[K];
      result := GUnit(A, LA, S1, S2, W[K], C );

      if result then
        Exit;

      if LA > N then
        for J:=0 to S3.Card-1 do
        begin
          if J = K then
            W[J].Free;

          if not Assigned(W[J]) then
            W[J] := TCollect.Create;
          for L:=N+1 to LA do
            if A[L].GetCount() = 1 then
              W[J].Add(A[L]);
        end;

      Inc(K);
      if K = S3.Card then
        K := 0;
    end;
  finally
    S1.Free;
    S2.Free;
    S3.Free;
  end;
end;

procedure Output(N: Integer);
var
  I, N1, N2: Integer;
begin
  N1 := HIST[HC].N1;
  N2 := HIST[HC].N2;
  Dec(HC);
  if N1 = 0 then
    writeln('Not proved')          
  else
  begin 
    MarkTree(N1, N);
    MarkTree(N2, N); 
    writeln('Proof: '); 
    for I:=N+1 to HC do
      if HIST[I].InProof then
        writeln(I:3, '(', HIST[I].N1, ',', HIST[I].N2, ',', HIST[I].ND, ')' );
    writeln('Contradiction: ', N1, ' ', N2);
  end;
end;

const
  A = 'A';
  D = 'D';
  F = 'F';
  H = 'H';
  L = 'L';
  P = 'P';
  X = 'X';
  Y = 'Y';
var
  I, N: Integer;
  Clauses: array of TClause;
  W: Poly;

procedure AddClause(const E: Poly);
begin
  Inc(N);
  Clauses[N] := TClause.Create(E);
end;

begin
  Num := 0;
  ArrConst := ['$1',  '$2',  '$3',  '$4',  '$5',  '$6',  '$7',  '$8' ];
  AX       := ['#X1', '#X2', '#X3', '#X4', '#X5', '#X6', '#X7', '#X8'];
  AY       := ['#Y1', '#Y2', '#Y3', '#Y4', '#Y5', '#Y6', '#Y7', '#Y8'];

  SetLength(Clauses, 100);
  N := 0;
  AddClause([['+',L,X,[F,X]]                       ]);
  AddClause([['-',L,X,X]                           ]);
  AddClause([['-',L,X,Y],['-',L,Y,X]               ]);
  AddClause([['-',D,X,[F,Y]],['+',L,Y,X]           ]);
  AddClause([['+',P,X],['+',D,[H,X],X]             ]);
  AddClause([['+',P,X],['+',P,[H,X]]               ]);
  AddClause([['+',P,X],['+',L,[H,X],X]             ]);
  AddClause([['-',P,X],['-',L,A,X],['+',L,[F,A],X] ]);

  SetLength(W, N);
  for I:=3 to N do
    W[I] := [1, 2];

  writeln('Source clauses:');

  for I:=1 to N do
  begin
    Clauses[I].Rename(AX);
    Clauses[I].Print;
  end;

  writeln('-------------------------------');
  writeln('Theorem proving');
  writeln('Unit binary resolution');
  writeln('-------------------------------');

  writeln('Theorem :');
  writeln('The set of prime numbers is unlimited');

  TPU(Clauses, W, N, 20);

  Output(N);
end.
